//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     ch.h
// Purpose:    Character classes
// Last Change:04.04.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 1998-2001 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	CH_INCLUDED
#define CH_INCLUDED

#ifdef	CH_MOD
#define	CH_EX
#else
#define	CH_EX	extern
#endif

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef BOOL_INCLUDED
#include "bool.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// CH_TABSIZE: Size of the character table, i.e. number of char values:
//-----------------------------------------------------------------------------

static const int CH_TABSIZE = 256;

//=============================================================================
// Exported Types:
//=============================================================================

//-----------------------------------------------------------------------------
// ch_t: Character classes.
//-----------------------------------------------------------------------------

typedef enum {
	CH_NULL,
	CH_EOF,
	CH_ATOM,
	CH_VAR,
	CH_MINUS,
	CH_DIGIT,
	CH_QUOTE,
	CH_FULLSTOP,
	CH_COMMA,
	CH_COLON,
	CH_SEMICOLON,
	CH_LPAREN,
	CH_RPAREN,
	CH_NOT,
	CH_AND,
	CH_OR,
	CH_LESS,
	CH_GREATER,
	CH_QUEST,
	CH_CMD,
	CH_UNUSED,
	CH_UNDEFINED
} ch_t;

//=============================================================================
// Auxillary Macros:
//=============================================================================

//-----------------------------------------------------------------------------
// CH_CHARCAST: How to get from characters to indices.
//-----------------------------------------------------------------------------

#define CH_CHARCAST(c)	((int) ((unsigned char)(c)))

//=============================================================================
// Class for Classifying Characters:
//=============================================================================

class ch_c {
	public:

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

	ch_c(void);

//-----------------------------------------------------------------------------
// init: Initialization of classification tables:
//-----------------------------------------------------------------------------

	static void	init(void);

//-----------------------------------------------------------------------------
// classify: Return the character class of a character.
//-----------------------------------------------------------------------------

	static inline	ch_t classify(char c)
		{ CHECK(Initialized,
			"ch_c::classify(not initialized)");
		  CHECK(CH_CHARCAST(c)<CH_TABSIZE,
			"ch_c::classify(index out of range)");
		  return(Classify[CH_CHARCAST(c)]); }

//-----------------------------------------------------------------------------
// is_alnum: Is this an alphanumeric character?
//-----------------------------------------------------------------------------

	static inline bool_t is_alnum(char c)
		{ CHECK(Initialized,
			"ch_c::is_alnum(not initialized)");
		  CHECK(CH_CHARCAST(c)<CH_TABSIZE,
			"ch_c::is_alnum(index out of range)");
		  return(IsAlnum[CH_CHARCAST(c)]); }

//-----------------------------------------------------------------------------
// is_space: Is this a space character (excluding newline)?
//-----------------------------------------------------------------------------

	static inline bool is_space(char c)
		{ return(c==' ' || c=='\t' || c=='\r'); }

//-----------------------------------------------------------------------------
// is_comment: Is this the comment character?
//-----------------------------------------------------------------------------

	static inline bool_t is_comment(char c)
		{ return(c=='%'); }

//-----------------------------------------------------------------------------
// is_newline: Is this the newline character?
//-----------------------------------------------------------------------------

	static inline bool_t is_newline(char c)
		{ return(c=='\n'); }

//-----------------------------------------------------------------------------
// is_quote: Is this the quote character?
//-----------------------------------------------------------------------------

	static inline bool_t is_quote(char c)
		{ return(c=='\''); }

//-----------------------------------------------------------------------------
// is_undersc: Is this the underscore character?
//-----------------------------------------------------------------------------

	static inline bool_t is_undersc(char c)
		{ return(c=='_'); }

//-----------------------------------------------------------------------------
// ord: Return position in "alphabet" (for sorting purposes).
//-----------------------------------------------------------------------------

	static inline int ord(char c)
		{ CHECK(Initialized,
			"ch_c::ord(not initialized)");
		  CHECK(CH_CHARCAST(c)<CH_TABSIZE,
			"ch_c::ord(index out of range)");
		  return(Ord[CH_CHARCAST(c)]); }

//-----------------------------------------------------------------------------
// xord: Position of second character equivalent to national character.
//-----------------------------------------------------------------------------

	static inline int xord(char c)
		{ CHECK(Initialized,
			"ch_c::xord(not initialized)");
		  CHECK(CH_CHARCAST(c)<CH_TABSIZE,
			"ch_c::xord(index out of range)");
		  return(Xord[CH_CHARCAST(c)]); }

//=============================================================================
// Private Auxillary Functions:
//=============================================================================

	private:

//-----------------------------------------------------------------------------
// enter: Enter character-class translation into table.
//-----------------------------------------------------------------------------

	static inline void	enter(char c, ch_t cl);

//=============================================================================
// Private Class Members:
//=============================================================================

//-----------------------------------------------------------------------------
// Initialized: This module was initialized.
//-----------------------------------------------------------------------------

	static bool_t	Initialized;

//-----------------------------------------------------------------------------
// Classify: Character class of each character.
//-----------------------------------------------------------------------------

	static ch_t	Classify[CH_TABSIZE];

//-----------------------------------------------------------------------------
// IsAlnum: Which characters are alphanumeric?
//-----------------------------------------------------------------------------

	static bool_t	IsAlnum[CH_TABSIZE];

//-----------------------------------------------------------------------------
// Ord: Position in "alphabet" (for sorting).
//-----------------------------------------------------------------------------

	static int	Ord[CH_TABSIZE];

//-----------------------------------------------------------------------------
// Xord: Position of second character equivalent to national character.
//-----------------------------------------------------------------------------

	static int	Xord[CH_TABSIZE];

//-----------------------------------------------------------------------------
// Xupper: Extended list of uppercase characters (including national ones).
//-----------------------------------------------------------------------------

	static str_t	Xupper;

//-----------------------------------------------------------------------------
// Xlower: Extended list of lowercase characters (including national ones).
//-----------------------------------------------------------------------------

	static str_t	Xlower;

//-----------------------------------------------------------------------------
// SzLig: The german "ss" character (special for alphabetic orders).
//-----------------------------------------------------------------------------

	static char	SzLig;

//-----------------------------------------------------------------------------
// AeUpper: The nordic "Ae" character (special for alphabetic orders).
//-----------------------------------------------------------------------------

	static char	AeUpper;

//-----------------------------------------------------------------------------
// AeLower: The nordic "ae" character (special for alphabetic orders).
//-----------------------------------------------------------------------------

	static char	AeLower;

//=============================================================================
// End of Class:
//=============================================================================

};

//=============================================================================
// End of File:
//=============================================================================

#endif

